#include "gridstorefunctions.h"
#include "gridstore.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <time.h>
#include <unistd.h>

#define MAX_OFFSET 12
#define PRESSURE_OFFSET 8

// Starts an infinite cycle that represents an espresso machine making espresso shots
// and sending timestamp records and updates into GridDB 
void espressoCycle(GSTimeSeries* espressoMachine,GSCollection* coffeeStore, char* serial_num,int shotRate, double standard_deviation,double average_temperature, double average_pressure,double failure_rate, double timeToCleanCycle){
	int currentCycle = 0; 
	int secondsPerShot = calculateSecondsPerShot(shotRate, standard_deviation, currentCycle); //Determine the wait time between espresso shots
	int secondsToCleanCycle = (int) (timeToCleanCycle * 3600); // Determine the amount of seconds until the next clean cycle
	int seed = 0;
	time_t timer;
	int offset,direction, pressure_direction, newShotRate;
	float pressure_offset, temperature,pressure;
	GSTimestamp basetime = gsCurrentTime();

	while(1){
		srand((unsigned) time(&timer));
		direction = rand();
		pressure_direction = rand();
		offset = rand() % (MAX_OFFSET + 1);
		pressure_offset = (float) (rand() % (PRESSURE_OFFSET + 1)) / 100.0; 
		if (direction % 2 == 1)
			offset = offset * -1;

		if (pressure_direction % 2 == 1)
			pressure_offset = pressure_offset * -1.0;


		pressure = average_pressure + pressure_offset;
		temperature = average_temperature + offset;

		
		if(currentCycle >= secondsToCleanCycle) {// Check if it is time to start a clean cycle 
			makeEspressoRecord(basetime,espressoMachine,coffeeStore, serial_num,temperature,pressure,failure_rate,seed,true); // Make an espresso with the eventType set to clean
			currentCycle = 0;
		}
		else
			makeEspressoRecord(basetime,espressoMachine,coffeeStore, serial_num,temperature,pressure,failure_rate,seed,false); // Make an espresso shot record
				
		basetime = gsAddTime(basetime,secondsPerShot, GS_TIME_UNIT_SECOND);
		seed += secondsPerShot;


		currentCycle += secondsPerShot; // Move to next cycle
		newShotRate = (int) adjustedShotsperHour(shotRate,standard_deviation, (unsigned) seed);
		secondsPerShot = calculateSecondsPerShot(newShotRate,standard_deviation, seed);

		sleep(secondsPerShot); // Wait the waiting time between shots
	}
}





int main(int argc, char** argv){
	if(argc < 3)
		printf("Need to have a serial number and configuration arguments\n");
	else {
		GSGridStore* gridstore = makeGridStore(HOST,PORT,CLUSTER_NAME,USER,PASSWORD); // Connect to GridDB
		GSTimeSeries* coffeeMachine = makeEspressoMachine(gridstore,argv[2]); // Make coffee machine TimeSeries container by its serial_number (containerName )
		GSCollection* coffeeStore;
		


		gsGetCollectionGeneral(gridstore,argv[1],&coffeeStore); // Get the coffee store the coffee machine belongs to
		

		if(coffeeMachine == NULL || coffeeStore == NULL)
			printf("The espresso machine or its accompanying coffee store failed to register\n");
		else {

			
			// Obtain machine settings from the user
			double failure_rate = atof(argv[3]); // Get failure rate
			int shotsPerHour = atoi(argv[4]); // Get machine's average rate of output
			double standard_deviation = atof(argv[5]) / 10.0;  // Get machine's deviation in how many shots it make per hour
			double clean_rate = atof(argv[6]); // Get machine's clean rate
			



			espressoCycle(coffeeMachine,coffeeStore,argv[2],shotsPerHour,standard_deviation,DEFAULT_TEMPERATURE,DEFAULT_PRESSURE,failure_rate,clean_rate); // Begin machine's espresso cycles
		}
		gsCloseGridStore(&gridstore,GS_TRUE);
	}
	return 0;
}