#include "gridstorefunctions.h"
#include "gridstore.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>



void preloadData(int days, GSTimeSeries* espressoMachine,GSCollection* coffeeStore, char* serial_num,int shotRate, double standard_deviation,double average_temperature, double average_pressure,double failure_rate, double timeToCleanCycle){
	GSTimestamp timeToInsert = gsAddTime(gsCurrentTime(),-1 * days, GS_TIME_UNIT_DAY); // Get the timestamp of 'days' days ago in GridDB timestamp format
	GSTimestamp timeToStopAt = gsCurrentTime();  // Preload until the current time
	int currentCycle = 0; // Number of cycles made 
	int secondsPerShot = calculateSecondsPerShot(shotRate,standard_deviation,currentCycle);  // Time between shots
	int secondsToCleanCycle = (int) (timeToCleanCycle * 3600);  // Time between clean cycles
	int currentSeconds  = 0; // Number of records
	int recordCount = 0;
	int newShotRate;

	// Generate and insert espresso timestamp records until the timestamp reaches the current time
	while(timeToInsert < timeToStopAt){
		if(currentCycle >= secondsToCleanCycle){
			makeEspressoRecord(timeToInsert,espressoMachine,coffeeStore,serial_num,average_temperature,average_pressure,failure_rate,currentSeconds,true);
			currentCycle = 0;
		}
		else
			makeEspressoRecord(timeToInsert,espressoMachine,coffeeStore,serial_num,average_temperature,average_pressure,failure_rate,currentSeconds,false);

		timeToInsert = gsAddTime(timeToInsert,secondsPerShot, GS_TIME_UNIT_SECOND);  // The next timestamp records will be the average wait time between shots as determined by the shot rate

		currentSeconds += secondsPerShot;
		currentCycle += secondsPerShot;
		

		if(newShotRate == 0)
			newShotRate = shotRate;
		recordCount++;
		newShotRate = (int) adjustedShotsperHour(shotRate,standard_deviation,(unsigned) currentCycle);
		secondsPerShot = calculateSecondsPerShot(newShotRate,standard_deviation,currentCycle);
		
	}

	
	printf("Preloaded %d records for the past %d days for coffee machine: %s\n",recordCount, days, serial_num);
}




int main(int argc, char** argv){
	if(argc < 3) {
		printf("Need to have a serial number and configuration arguments\n");
		return 1;
	}

	GSGridStore* gridstore = makeGridStore(HOST,PORT,CLUSTER_NAME,USER,PASSWORD); // Connect to GridDB
	GSTimeSeries* coffeeMachine = makeEspressoMachine(gridstore,argv[2]); // Make coffee machine TimeSeries container by its serial_number (containerName )


	GSCollection* coffeeStore;
	gsGetCollectionGeneral(gridstore,argv[1],&coffeeStore); // Get the coffee store the coffee machine belongs to
	if(coffeeMachine == NULL || coffeeStore == NULL) {
		printf("The espresso machine or its accompanying coffee store failed to register\n");
		return 1;
	}

	// Obtain machine settings from the user
	double failure_rate = atof(argv[3]); // Get failure rate
	int shotsPerHour = atoi(argv[4]); // Get machine's average rate of output
	double standard_deviation = atof(argv[5]) / 10.0;  // Get machine's deviation in how many shots it make per hour
	double clean_rate = atof(argv[6]); // Get machine's clean rate
	int days_to_preload = atoi(argv[7]); // Get amount of days to preload from

	if (days_to_preload > 0)
		// Preload data that would represent coffee machine behavior from the past
		preloadData(days_to_preload,coffeeMachine,coffeeStore,argv[2],shotsPerHour,standard_deviation,DEFAULT_TEMPERATURE,DEFAULT_PRESSURE,failure_rate,clean_rate); 

	gsCloseGridStore(&gridstore,GS_TRUE); // Clean up and close connection to GridDB 

	return 0;
}
