#include "gridstore.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define SAMPLE_CHAIN "oc_sample_2"
#define DEFAULT_NUM_STORES 3
#define DEFAULT_NUM_MACHINES 5

void dropDefaultChain(GSGridStore* gridstore){
	gsDropCollection(gridstore,SAMPLE_CHAIN);
	printf("Dropped coffee chain %s\n",SAMPLE_CHAIN);
	int i, j;
	char store_ids[6];
	char serial_nums[8];
	for(i = 1; i < DEFAULT_NUM_STORES + 1; i++){
		sprintf(store_ids,"A_%d",i);
		gsDropCollection(gridstore,store_ids);
		printf("Dropped coffee store %s\n",store_ids);
		for(j = 1; j < DEFAULT_NUM_MACHINES + 1; j++){
			sprintf(serial_nums,"M_%d_%d",i,j);
			gsDropTimeSeries(gridstore,serial_nums);
			printf("Dropped coffee machine %s\n",serial_nums);
		}
	}

}


// In cases where the user may wish to reset all the data for a coffee chain this file can be used to wipe the data 
// Coffee chain container and all the other containers (coffee stores and coffee machines) that it corresponds to
int main(int argc, char** argv){
	if(argc < 7)
		printf("Need a coffee chain name and GridDB connection settings\n");
	else {
		GSGridStore* gridstore;
		const GSPropertyEntry properties[] = {
			{"notificationAddress",argv[2]},
			{"notificationPort",argv[3]},
			{"clusterName",argv[4]},
			{"user",argv[5]},
			{"password",argv[6]}
		}; // Set Connection Properties
		size_t propCount = sizeof(properties) / sizeof(*properties);
		gsGetGridStore(gsGetDefaultFactory(),properties,propCount,&gridstore); // Connect to GridDB cluster
		if(strcmp(argv[1],SAMPLE_CHAIN) == 0){
			printf("Removing default Sample chain\n");
			dropDefaultChain(gridstore);
		}
		else {
			GSCollection* coffee_chain; // Coffee Chain to remove
			printf("You have chosen to remove coffee chain: %s\n",argv[1]);





			gsGetCollectionGeneral(gridstore,argv[1],&coffee_chain);
			if(coffee_chain == NULL){
				printf("The coffee chain does not exist\n"); // If the coffee chain is not in GridDB exit immediately
				gsCloseGridStore(&gridstore,GS_TRUE);
				return 1;
			}

			GSQuery* query;
			gsQuery(coffee_chain,"select *",&query);
			GSRowSet* rowSet;
			gsFetch(query,GS_FALSE,&rowSet); // Retrieve and fetch all coffee stores from the coffee chain and remove them
			GSRow* row;
			


			while (gsHasNextRow(rowSet)){
				const GSChar* container_name; // Store ID: of Coffee Store
				GSChar buffer[GS_TIME_STRING_SIZE_MAX]; // At a certain point the original coffee store id is lost so a copy is needed so it can be dropped later from
													// GridDB
				GSQuery* innerQuery; 
				GSRowSet* innerRs;
				GSRow* storeEntry;
				GSCollection* coffee_store;
			

				gsCreateRowByContainer(coffee_chain,&row); // Create a row entry for a Coffee Store
				gsGetNextRow(rowSet,row); // Obtain coffee store information
				gsGetRowFieldAsString(row,0,&container_name); // Obtain store id


				gsGetCollectionGeneral(gridstore,container_name,&coffee_store); // Find a Coffee Store container with the corresponding Store ID as its name
				strcpy(buffer,container_name); // Make a copy of the name so the container can be dropped later
				gsQuery(coffee_store,"select *",&innerQuery); // Get all coffee machines in store
				gsFetch(innerQuery,GS_FALSE,&innerRs); // Fetch results



				while(gsHasNextRow(innerRs)){
					gsCreateRowByContainer(coffee_store,&storeEntry); // 
					gsGetNextRow(innerRs,storeEntry); // Obtain coffee machine information
					const GSChar* serial_num; // Machine Number of the TimeSeries container
					gsGetRowFieldAsString(storeEntry,0,&serial_num); // Obtain machine's serial number
					printf("Removing timeseries: %s\n",serial_num);
					gsDropTimeSeries(gridstore,serial_num); // Drop the corresponding TimeSeries
				}




				printf("Removing Collection %s\n",buffer);
				gsDropCollection(gridstore,buffer); // Now that all the coffee machine containers are removed, the Coffee store can be removed as well
			}
			


			gsDropCollection(gridstore,argv[1]); // With all coffee stores gone, the coffee chain can be removed
			printf("Removed Collection: %s\n",argv[1]);
		}
		gsCloseGridStore(&gridstore,GS_TRUE); // Exit from GridDB
	}
	return 0;
}
