#!/usr/bin/python
import sys, traceback
import griddb_python_client
from bokeh.application import Application
from bokeh.server.server import Server
from bokeh.application.handlers import FunctionHandler
from datetime import date, timedelta, datetime
import time
import GridstoreMethods
from bokeh.plotting import figure
from bokeh.layouts import widgetbox, column, row
from bokeh.models import ColumnDataSource, Div, Range1d, DatetimeTickFormatter, DataRange1d, ResetTool, HelpTool, Legend, HoverTool, ZoomOutTool, ZoomInTool
from bokeh.models.widgets import Select
from bokeh.models.callbacks import CustomJS

num_procs = 4


INITIAL_DAYS = 1

host = "localhost"
port = "5006"

chain_name = "oc_sample_2"

db_host = "239.0.0.1"
db_port = "31999"
db_cluster_name = "defaultCluster"
db_user = "admin"
db_password = "admin"


LINE_COLORS = ["red","blue","green","purple","orange","yellow","black","#800000","#CD5C5C","#DC143C", "#8B0000", "#DB7093", "#FF1493", 
"#FF4500","#FFFF00", "#BDB76B","#FF00FF", "#663399", "#8B008B" ,"#4B0082", "#32CD32" ,"#006400" ,"#008B8B", "#4682B4" ,"#00BFFF"]

COLOR_COUNT = 25

DEFAULT_INTERVAL = 1000 * 60 * 60 * 24

#Default Plot Tools Provided Bokeh Library
TOOLS = "save,xpan,box_zoom,undo,redo,xwheel_zoom"

chain_name = "oc_sample_2"
hours = 0



if len(sys.argv) > 1:
	host = sys.argv[1]
	if host == "0.0.0.0":
		host = "*.*.*.*"
if len(sys.argv) > 2:
	port = sys.argv[2]
if len(sys.argv) > 3:
	num_procs = int(sys.argv[3])
if len(sys.argv) > 4:
	chain_name = sys.argv[4]
if len(sys.argv) > 5:
	db_host = sys.argv[5]
if len(sys.argv) > 6:
	db_port = sys.argv[6]
if len(sys.argv) > 7:
	db_cluster_name = sys.argv[7]
if len(sys.argv) > 8:
	db_user = sys.argv[8]
if len(sys.argv) > 9:
	db_password = sys.argv[9]


displayed_host = host
if displayed_host == "*.*.*.*":
	displayed_host = "0.0.0.0"

selectedVal = ""

serial_numbers = []

line_sources = []
temperature_sources = []


SHOT_RATE = "SHOT_RATE"
TEMPERATURE = "TEMPERATURE"

socket_origins = ["{0}:{1}".format(host,port)]
if host == "127.0.0.1":
	socket_origins.append("localhost:{0}".format(port))
if host == "localhost":
	socket_origins.append("127.0.0.1:{0}".format(port))


def get_past_readings(gridstore,line_source,serial_number,timestamps,setting):
	"""Obtains readings from a coffee machine for a certain amount of timestamps and adds its data to a plot line's data source. """
	coffee_machine = gridstore.get_container(serial_number)
	for timestamp in timestamps:
		y_val = 0
		past_date = datetime.fromtimestamp(int(timestamp) / 1000 )
		if setting == SHOT_RATE:
			y_val = GridstoreMethods.get_instantaneous_rate(coffee_machine,timestamp)
		elif setting == TEMPERATURE:
			y_val = GridstoreMethods.get_recent_readings(gridstore,serial_number,int(timestamp))["temperature"]
		line_source.data["x"].insert(0,past_date)
		line_source.data["y"].insert(0,y_val)


def get_past_shot_rates(gridstore,serial_number,days):
	"""Fetch all past timestamps for successful coffee machine shots for collecting past shot rates and timestamps"""
	x = []
	y = []
	try:
		coffee_machine = gridstore.get_container(serial_number)
		tql = "select * where event_type=1 and timestamp >= TIMESTAMPADD(DAY,NOW(),{0}) order by timestamp asc".format(days * -1)
		query = coffee_machine.query(tql)
		minimum = GridstoreMethods.get_time(coffee_machine,"MIN")
		rowSet = query.fetch(False)
		row = coffee_machine.create_row()
		while rowSet.has_next():
			rowSet.get_next(row)
			timestamp = row.get_field_as_timestamp(0)
			shotRate = GridstoreMethods.get_instantaneous_rate(coffee_machine,int(timestamp))
			x.append(datetime.fromtimestamp(int(timestamp) / 1000))
			y.append(shotRate)
	except:
		print("Error fetching shot rates from coffee machine: " + serial_number)
		traceback.print_stack()
	return x, y

def get_past_temperatures(gridstore,serial_number,days):
	"""Fetch all past timestamps for a certain amount of days for collecting timestamps and temperatures"""
	x = []
	y = []
	try:
		coffee_machine = gridstore.get_container(serial_number)
		tql = "select * where timestamp <= NOW() and timestamp >= TIMESTAMPADD(DAY,NOW(),{0}) order by timestamp asc".format(days * -1)
		minimum = GridstoreMethods.get_time(coffee_machine,"MIN")
		query = coffee_machine.query(tql)
		rowSet = query.fetch(False)
		row = coffee_machine.create_row()
		while rowSet.has_next():
			rowSet.get_next(row)
			timestamp = row.get_field_as_timestamp(0)
			temperature = row.get_field_as_float(2)
			x.append(datetime.fromtimestamp(int(timestamp) / 1000))
			y.append(temperature)
	except:
		print("Error fetching temperatures from machine " + serial_number)
		traceback.print_stack()
	return x, y

def create_timeseries(gridstore,store_id,serial_nums,callback):
	"""Create an empty timeseries plot and a data source that links """

	line_sources = []
	items = []
	timeseries = figure(name="Toggled Plot",title = "Machines in Coffee Store " + store_id,x_axis_type = 'datetime',x_axis_label = "Current Time(Calendar)" , x_range=DataRange1d(follow="end",range_padding=0.0,default_span=DEFAULT_INTERVAL ,follow_interval=DEFAULT_INTERVAL),
	y_range = Range1d(20,100), y_axis_label="Shots Per Hour",plot_width=1000,plot_height=600,tools=TOOLS,toolbar_location="above")
	for idx in range(len(serial_nums)):
		x, y = get_past_shot_rates(gridstore,serial_nums[idx],INITIAL_DAYS)
		shot_line = timeseries.line("x","y",source = ColumnDataSource(data = dict(x = x, y = y)), color=LINE_COLORS[idx % COLOR_COUNT])
		shot_line_source = shot_line.data_source ### Data source of timeseries line that can be updated and changed in callback functions
		items.append(("Coffee Machine: " + serial_nums[idx],[shot_line]))
		line_sources.append(shot_line_source)

	## Add legend and its layout ##
	legend = Legend(items = items, location = (30,-30))
	timeseries.add_layout(legend,'right')


	xformatter = DatetimeTickFormatter(hours = ["%a %H:%M:%S"],minutes = ["%H:%M:%S"],seconds = ["%H:%M:%S"]) ### Have datetimes appear as formatted calendar dates and times
	timeseries.xaxis.formatter = xformatter
	timeseries.xaxis.major_label_orientation = "vertical"   ### Rotate x ticks so they can be easier to read
	
	## Set plot styling settings ##
	timeseries.min_border_bottom = 100
	timeseries.title.align = 'left'
	
	timeseries.x_range.on_change('start',callback) ## Add pan callback event handler
	
	## Add extra tools to plot toolbar ##
	timeseries.add_tools(ResetTool(reset_size=True))
	timeseries.add_tools(HelpTool(help_tooltip="Learn more about GridDB at their official website",redirect="https://griddb.net/en"))
	timeseries.add_tools(ZoomInTool(dimensions = 'width', factor = 0.25))
	timeseries.add_tools(ZoomOutTool(dimensions = 'width', factor = 0.25))


	return timeseries,line_sources   




def create_temperature_plot(gridstore,store_id,serial_nums,callback):
	"""Create an empty timeseries plot for plotting temperature data"""
	timeseries = figure(name="Temperature Timeseries",title = "Machines in Coffee Store " + store_id, x_axis_type="datetime",x_axis_label="Current Time(Calendar)",y_range=Range1d(160,200),
	x_range=DataRange1d(follow='end',range_padding=0.0,default_span=DEFAULT_INTERVAL,follow_interval = DEFAULT_INTERVAL),y_axis_label="Current Temperature (Fahrenheit)", plot_width = 1000,plot_height=800,tools=TOOLS,toolbar_location='above')
	line_sources = []
	
	items = []


	for idx in range(len(serial_nums)):
		x,y = get_past_temperatures(gridstore,serial_nums[idx],INITIAL_DAYS)
		shot_line = timeseries.line("x","y",source = ColumnDataSource(dict(x = x, y = y)),color = LINE_COLORS[idx % COLOR_COUNT])
		shot_line_source = shot_line.data_source
		items.append(("Coffee Machine: " + serial_numbers[idx],[shot_line]))
		line_sources.append(shot_line_source)


	## Add legend and its layout to plot ##
	legend = Legend(items = items, location = (30,-30),label_width = 200)
	timeseries.add_layout(legend,'right')

	xformatter = DatetimeTickFormatter(hours = ["%a %H:%M:%S"],minutes = ["%H:%M:%S"],seconds = ["%H:%M:%S"]) ### Have datetimes appear as formatted calendar dates and times
	timeseries.xaxis.formatter = xformatter
	timeseries.xaxis.major_label_orientation = 'vertical'	
	
	## Add plot style properties ##
	timeseries.min_border_bottom = 100
	timeseries.title.align = 'left'


	timeseries.x_range.on_change('start',callback) ## Add pan callback event handler


	hover = HoverTool(tooltips = [("Date: ","@x{%F}"),("Temperature: ","$y &deg; F")],formatters=dict(x='datetime'))
	timeseries.add_tools(hover)
	## Add extra tools to plot toolbar ## 
	timeseries.add_tools(ResetTool(reset_size=True))
	timeseries.add_tools(HelpTool(help_tooltip="Learn more about GridDB at their Official Website", redirect="https://griddb.net/en/"))
	timeseries.add_tools(ZoomInTool(dimensions = 'width', factor = 0.25))
	timeseries.add_tools(ZoomOutTool(dimensions = 'width', factor = 0.25))
	return timeseries, line_sources



def modify_doc(doc):
	"""Base Function for creating the webpage's document template and setting its callback functions"""
	global selectedVal, serial_numbers, line_sources, temperature_sources
	gridstore = GridstoreMethods.fetchGridStore(db_host,db_port,db_cluster_name,db_user,db_password)
	#### Webpage Methods
	def create_callback(div,select):
		"""Have the text in the caption div change when the selected value in the options menu changes"""
		return CustomJS(args=dict(div=div,select=select),code='''
			div.text = "Below is the plot for coffee store " + select.value; ''')


	def xrangeCallback(attr,old,new):
		"""Callback function to plot past shot rates when user pans left"""
		global line_sources
		if type(old) != type(None) and type(new) != type(None):
			min_date = line_sources[0].data["x"][0]
			diff = 1
			date_diff = timedelta(1)
			timestamp = min_date
			if type(min_date) == float or type(min_date) == int:
				diff = new - min_date
			else:
				new_date = datetime.fromtimestamp(int(new) / 1000)
				date_diff = new_date - min_date
				timestamp = time.mktime(min_date.timetuple()) * 1000.0
			## If the user pans to a date that is in the past and not plotted on the graph fetch the past data and plot it ##
			if (diff < 0 or date_diff < timedelta(0)):

				for index in range(len(line_sources)):
					timestamps = GridstoreMethods.fetch_timestamps_betw(gridstore,serial_numbers[index],int(new),int(timestamp), filtered = True)
					get_past_readings(gridstore,line_sources[index],serial_numbers[index],timestamps,SHOT_RATE)



	def findPreviousTemperatures(attr,old,new):
		"""Callback function to plot past temperature when user pans left"""
		global temperature_sources

		if type(old) != type(None) and type(new) != type(None):
			min_date = temperature_sources[0].data["x"][0]
			diff = 1
			date_diff = timedelta(1)
			timestamp = min_date
			if type(min_date) == float or type(min_date) == int:
				diff = new - min_date 
			else:
				new_date = datetime.fromtimestamp(int(new) / 1000)
				date_diff = new_date - min_date
				timestamp = time.mktime(min_date.timetuple()) * 1000.0
				### If the user pans to an earlier date fetch a past data from the start of the xrange and plot it on the graph ###
			
			if (diff < 0 or date_diff < timedelta(0)):
				for index in range(len(temperature_sources)):
					timestamps = GridstoreMethods.fetch_timestamps_betw(gridstore,serial_numbers[index],int(new),int(timestamp))
					get_past_readings(gridstore,temperature_sources[index],serial_numbers[index],timestamps,TEMPERATURE)


	## Connect to Main Coffee Chain
	
	coffee_chain = gridstore.get_container(chain_name)
	

	store_ids = filter(lambda store: len(store) > 0, GridstoreMethods.get_store_ids(coffee_chain))  ### Get all coffee stores in the coffee chain

	div = Div(text="Below is the plot for the coffee store " + store_ids[0],height=100,width = 1000)  #### Currently Displayed Store


	serial_numbers = filter(lambda serial_num: len(serial_num) > 0, GridstoreMethods.get_serial_nums(gridstore,store_ids[0]))


	### Add initial shot rate and temperature timeseries graphs ###
	timeseries_plots,line_sources = create_timeseries(gridstore,store_ids[0],serial_numbers,xrangeCallback)

	temperature_plot, temperature_sources = create_temperature_plot(gridstore,store_ids[0],serial_numbers,findPreviousTemperatures)



	select = Select(title = "Coffee Store", value = store_ids[0], options=store_ids)  ### Create an option menu where users can choose to display timeseries of the store of their choosing
	



	select.callback = create_callback(div,select)  ### Create Javascript callback so text in HTML div text can be linked to the selected store in the option menu

					### Layout ####
					  ### Top ###
	# 1. Store Option Menu               #
	# 2. Caption Text (HTML div)         #
	# 3. TimeSeries Plot of Coffee Store #
	# 4. Temperature Plot                #
	layout = column(widgetbox(select),div, timeseries_plots, temperature_plot, name="mainLayout")

	doc.add_root(layout)


	def update():
		"""Periodic callback function used by the Bokeh Server"""
		global selectedVal,serial_numbers,line_sources,temperature_sources
		### Obtain Layouts and Timeseries Plots ###
		mainLayout = doc.get_model_by_name("mainLayout")
		timeseries = doc.get_model_by_name("Toggled Plot")
		temperature_timeseries = doc.get_model_by_name("Temperature Timeseries")
		subLayouts = mainLayout.children


		if selectedVal != select.value:
			selectedVal = select.value ### Change currently selected coffee store
			
			## Remove components of old coffee store
			subLayouts.remove(timeseries)
			subLayouts.remove(temperature_timeseries)	

			serial_numbers = filter(lambda serial_num: len(serial_num) > 0, GridstoreMethods.get_serial_nums(gridstore,str(selectedVal))) ### Obtain new list of coffee machines
			
			timeseries_plots, line_sources = create_timeseries(gridstore,str(selectedVal),serial_numbers,xrangeCallback) ## Get new timeseries plots
			temperature_plot, temperature_sources = create_temperature_plot(gridstore,str(selectedVal), serial_numbers, findPreviousTemperatures)
			
			## Insert new visualisation components
			subLayouts.insert(2,timeseries_plots)
			subLayouts.insert(3,temperature_plot)


		for index in range(len(serial_numbers)):
			try:
				coffee_machine = gridstore.get_container(serial_numbers[index]) ## Obtain individual coffeee machine
				
				needsShotUpdate = False
				needsTemperatureUpdate = False

				last_timestamp = 0
				new_rate_times = []
				new_temperatures_times = []

				if len(line_sources[index].data["x"]) > 0:
					last_timestamp = line_sources[index].data["x"][-1]
					if type(last_timestamp) == int or type(last_timestamp) == float:
						last_timestamp = int(last_timestamp)
					else:
						last_timestamp = int(time.mktime(last_timestamp.timetuple()) * 1000)
					new_rate_times = GridstoreMethods.fetch_timestamps_betw(gridstore,serial_numbers[index],last_timestamp,GridstoreMethods.Timestamp_current(), filtered = True)
					new_temperatures_times = GridstoreMethods.fetch_timestamps_betw(gridstore,serial_numbers[index],last_timestamp,GridstoreMethods.Timestamp_current())

				for idx in range(len(new_rate_times) -1 , -1, -1):
					new_avg = GridstoreMethods.get_instantaneous_rate(coffee_machine,new_rate_times[idx])
					line_sources[index].data["x"].append(datetime.fromtimestamp(int(new_rate_times[idx]) / 1000 ))
					line_sources[index].data["y"].append(new_avg)


				for idx in range(len(new_temperatures_times)-1, -1, -1):
					new_temperature = GridstoreMethods.get_recent_readings(gridstore,serial_numbers[index],int(new_temperatures_times[idx]))["temperature"]
					temperature_sources[index].data["x"].append(datetime.fromtimestamp(int(new_temperatures_times[idx]) / 1000))
					temperature_sources[index].data["y"].append(new_temperature)

					### Render Updates ###
				line_sources[index].trigger('data',line_sources[index].data,line_sources[index].data)
				temperature_sources[index].trigger('data',temperature_sources[index].data,temperature_sources[index].data)

				
			except:
			 	print("Error fetching coffee machine")


	doc.add_periodic_callback(update,4000) ## Add callback to document with callback rate


application = Application(FunctionHandler(modify_doc))

## This line of code is for bokeh 12.7
server = Server({'/':application}, allow_websocket_origin = socket_origins, port = int(port), num_procs = num_procs)

## This line of code will work for bokeh 12.6
#server = Server({'/':application},allow_websocket_origin = ["{0}:{1}".format(host,port)], port = int(port), num_proc = num_procs)
server.start()


if __name__ == "__main__":
	print("Starting a GridDB Bokeh application at http://{0}:{1} supported by {2} processes.".format(displayed_host,port,num_procs))
	print("Open a browser and type http://{0}:{1} into the address bar".format(displayed_host,port))
	# server.io_loop.add_callback(server.show,'/')
	# server.io_loop.start()
	server.run_until_shutdown()
